
import os
import re
from .constants import *
from .providers import KeyValueStoreProvider, Providers
from .logger import Logger
from .exceptions import DSOException
from .stages import Stages
from .appconfig import AppConfig, ContextSource


class ConfigProvider(KeyValueStoreProvider):
    def list(self, service, uninherited, filter):
        raise NotImplementedError()
    def set(self, key, value, service):
        raise NotImplementedError()
    def get(self, key, revision, service):
        raise NotImplementedError()
    def history(self, key, service):
        raise NotImplementedError()
    def unset(self, key, service):
        raise NotImplementedError()


class ConfigService():

    def list(self, service, uninherited=False, filter=None):
        self.service = service
        provider = Providers.ConfigProvider()
        Logger.debug(f"Listing configurations: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = provider.list(service=service, uninherited=uninherited, filter=filter)
        return response


    def set(self, key, value, service):
        self.service = service
        if not key: key = os.path.basename(filepath)
        # self.validate_key(key)
        provider = Providers.ConfigProvider()
        Logger.debug(f"Setting configuration '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.set(key=key, value=value, service=service)


    def get(self, key, service):
        self.service = service
        provider = Providers.ConfigProvider()
        Logger.debug(f"Getting configuration '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.get(key=key, service=service)


    # def history(self, key):
    #     self.config = config
    #     provider = Providers.ConfigProvider()
    #     Logger.debug(f"Getting the history of configuration '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
    #     return provider.history(service=service, key)


    def uset(self, key, service):
        self.service = service
        provider = Providers.ConfigProvider()
        Logger.debug(f"Unsetting configuration '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.unset(key=key, service=service)


Configs = ConfigService()



import logging
import colorama
import coloredlogs
from .settings import *


FATAL = 0 ### logs only fatal errors (not handled)
ERROR = 1 ### logs only errors
WARNING = 2 ### logs also warnings
INFO = 3 ### logs also information
DETAIL = 4 ### logs also detailed information
DEBUG = 5 ### logs also debug information
EXCEPTION = 6 ### logs also unhandled exceptions
PACKAGE = 7 ### also sets built-in logger level to logging.INFO
FULL = 8 ### also sets built-in logger level to logging.DEBUG


FIELD_STYLES = dict(
    asctime=dict(color='green', faint=True, bold=BOLD_LOGS),
    levelname=dict(color='cyan', faint=True, bold=BOLD_LOGS),
)

LEVEL_STYLES = dict(
    fatal=dict(color='red', bold=BOLD_LOGS),
    error=dict(color='red', bold=BOLD_LOGS),
    warning=dict(color='yellow', bold=BOLD_LOGS),
    info=dict(color='white', bold=BOLD_LOGS),
    detail=dict(color='white', faint=True, bold=BOLD_LOGS),
    debug=dict(color='white', faint=True, bold=BOLD_LOGS),
)

### to control logging verbosity of external libraries
verbosity_map = {
    FATAL : logging.FATAL,
    ERROR : logging.ERROR,
    WARNING : logging.WARNING,
    INFO : logging.WARNING,
    DETAIL : logging.WARNING,
    DEBUG : logging.WARNING,
    EXCEPTION : logging.WARNING,
    PACKAGE : logging.INFO,
    FULL : logging.DEBUG,
}

level_name_map = {
    'FATAL': FATAL,
    'ERROR': ERROR,
    'WARNING': WARNING,
    'INFO': INFO,
    'DETAIL': DETAIL,
    'DEBUG': DEBUG,
    'EXCEPTION': EXCEPTION,
    'PACKAGE': PACKAGE,
    'FULL': FULL,
}

def get_user_data_dir(appname):
    from pathlib import Path
    import os
    if WIN_PLATFORM:
        import winreg
        key = winreg.OpenKey(
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
        )
        dir_,_ = winreg.QueryValueEx(key, "Local AppData")
        result = Path(dir_).resolve(strict=False)
    elif MAC_PLATFORM:
        result = Path('~/Library/Application Support/').expanduser()
    elif NIX_PLATFORM:
        result = Path(os.getenv('XDG_DATA_HOME', "~/.local/share")).expanduser()
    else:
        raise NotImplementedError
    os.makedirs(result.joinpath(appname, 'logs'), exist_ok=True)
    return result.joinpath(appname, 'logs')


def get_log_file_name():
    from datetime import datetime
    filename = datetime.now().strftime('%Y-%m-%d.log')
    return os.path.join(get_user_data_dir('dso'), filename) 


class LoggerWrapperClass():

    # logging.addLevelName(DETAIL, 'DETAIL')
    def __detail(self, message, *args, **kws):
        if self.isEnabledFor(DETAIL):
            # Yes, logger takes its '*args' as 'args'.
            self._log(DETAIL, message, args, **kws) 
    # logging.Logger.detail = self.__detail

    def get_log_format(self):
        log_format = ''
        if TIMESTAMP_LOGS: log_format += '%(asctime)s'
        if LABEL_LOG_LEVELS: log_format += ' [%(levelname)-5s]'
        log_format += ' %(message)s'
        return log_format

    def __init__(self, level):
        self.level = level
        logging.basicConfig(format=self.get_log_format(), datefmt=LOG_TIME_FORMAT, level=self.mapped_level, force=True)
        ### rename critical to fatal
        logging.addLevelName(logging.CRITICAL, 'FATAL')
        ### rename WARNING to WARN
        logging.addLevelName(logging.WARNING, 'WARN')
        ### add new level DETAIL 
        logging.addLevelName(level_name_map['DETAIL'], 'DETAIL')
        ### set formats and root level, and force recreate any exisitng handler

        self.logger = logging.getLogger('dso')
       
        ### set level of the logger to DEBUG to catch all
        self.logger.setLevel(logging.DEBUG)
        self.logger.detail = self.__detail

        #### Add file handler to capture all logs, regardless of verbosity level
        fh = logging.FileHandler(get_log_file_name())
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter(self.get_log_format(), LOG_TIME_FORMAT))
        self.logger.addHandler(fh)
        ### also add it to the root to capture messages from other libraries as well
        self.logger.root.addHandler(fh)

        #### Add console handler
        if COLORED_LOGS:
            ### set level always to maximum (as the level is controlled by mapped_level), and attach it to root logger
            coloredlogs.install(logger=logging.root, 
                                level='DEBUG',
                                fmt=self.get_log_format(),
                                level_styles=LEVEL_STYLES,
                                field_styles=FIELD_STYLES,
                                )
        else:
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            ch.setFormatter(logging.Formatter(self.get_log_format(), LOG_TIME_FORMAT))
            self.logger.addHandler(ch)
            ### also add it to the root to capture messages from other libraries as well
            self.logger.root.addHandler(ch)


    ### maps dso log verbosity to internal verbosity
    def _map_level(self, level):
        return verbosity_map[level]

    @property
    def mapped_level(self):
        return self._map_level(self.level)

    def set_verbosity(self,level):
        self.level = level
        logging.root.setLevel(self.mapped_level)
        # for handler in self.logger.handlers:
        #     if isinstance(handler, type(logging.StreamHandler())):
        #         handler.setLevel(self.mapped_level)

    def increase_verbosity(self):
        changed = self.level < FULL
        if changed:
            self.set_verbosity(self.level + 1)
        return changed

    def decrease_verbosity(self):
        changed = self.level > FATAL
        if changed:
            self.set_verbosity(self.level - 1)
        return changed

    def fatal(self, msg, force=True):
        self.logger.propagate = force or self.level >= FATAL
        self.logger.fatal(msg)

    def error(self, msg, force=True):
        self.logger.propagate = force or self.level >= ERROR
        self.logger.error(msg)

    def warn(self, msg, force=False):
        self.logger.propagate = force or self.level >= WARNING
        self.logger.warning(msg)

    def info(self, msg, force=False):
        self.logger.propagate = force or self.level >= INFO
        self.logger.info(msg)

    ### FIXME
    def detail(self, msg, force=False):
        self.logger.propagate = force or self.level >= DETAIL
        # self.logger.log(DETAIL, msg)
        # self.logger.log(logging.INFO, msg)
        self.logger.detail(self.logger, msg)

    def debug(self, msg, force=False):
        self.logger.propagate = force or self.level >= DEBUG
        self.logger.debug(msg)

Logger = LoggerWrapperClass(level_name_map[DEFAULT_LOG_LEVEL])




import sys
import os
import ipaddress
import json
import yaml
import re
from logger import Logger
from settings import *
from exceptions import DSOException

UNNAMED_SUBNET = 'N/A'

def ipv4_to_binary(ip, sep="."):
    return  sep.join(map(str,["{0:08b}".format(int(x)) for x in ip.split(".")]))


class IPv4AddressX(ipaddress.IPv4Address):
    def to_binary(self, sep="."):
        """Represent IPv4 as 4 blocks of 8 bits."""
        return sep.join(f"{i:08b}" for i in self.packed)

    @classmethod
    def from_binary_repr(cls, binary_repr: str):
        """Construct IPv4 from binary representation."""
        # Remove anything that's not a 0 or 1
        i = int(re.sub(r"[^01]", "", binary_repr), 2)
        return cls(i)


def execute_subnet_plan(subnet_plan, summary=True):

    def add_partition_subnets(partitions, partition_ix, supernet, result: dict):

        def is_subnet_selected(supernet, subnet, selects):

            if not subnet.subnet_of(supernet):
                return False

            if not isinstance(selects, list): selects = [str(selects)]

            subnetIP = IPv4AddressX(subnet.network_address).to_binary(sep='')

            result = True
            for select in selects:
                ### inverse select is used?
                if select.startswith('!'):
                    select = select[1:]
                    inverse = True
                else:
                    inverse = False

                ### binary format
                if select.startswith('0b'):
                    select = select[2:]
                    m = re.match(r'^([x]*)([01]*)$', select)
                    if not m:
                        raise DSOException(f"Invalid value for 'select': {select}")
                    offset = len(m.group(1))
                    select = m.group(2)
                else:
                    raise NotImplementedError

                # print(f"1={supernetIP[:supernet.prefixlen]},{supernetIP[supernet.prefixlen:]}")
                # print(f"2={subnetIP[:supernet.prefixlen]},{subnetIP[supernet.prefixlen:]}")
                # print(f"3={subnetIP[supernet.prefixlen + offset:]}")
                # print(f"offset={offset}, select={select}, S={subnetIP[supernet.prefixlen + offset:]}")
                starting = subnetIP[supernet.prefixlen + offset:].startswith(select)
                selecetd = starting if not inverse else not(starting)

                result = result and selecetd

            return result

        ### if last partition, add host count
        if partition_ix == len(partitions):
            hosts = list(supernet.hosts())
            result['Hosts'] = {
                'Count': len(hosts),
                'IP': f"{hosts[0]}/32-{hosts[-1]}/32"
            }
            return

        partition = partitions[partition_ix]
        if partition['bits'] > 0:
            subnets = list(supernet.subnets(partition['bits']))
            result[partition['name']] = {
                'Count': 0, ### to be incremented while adding subnets
                'Ranges': []
            }
            range_first_subnet = None
            range_last_subnet = None
            for i, subnet in enumerate(subnets):
                Logger.debug(f"{partition['name']} {i}: {subnet}")
                if 'select' in partition:
                    if not is_subnet_selected(supernet, subnet, partition['select']): 
                        ### add the last subnet in the range if in summary mode
                        if summary and not range_last_subnet is None:
                            result[partition['name']]['Ranges'][-1]['Subnets'].append({
                                'CIDR' : range_last_subnet.with_prefixlen,
                                'Name' : partition['labels'].get(str(i), UNNAMED_SUBNET),
                            })
                            add_partition_subnets(partitions, partition_ix+1, range_last_subnet, result[partition['name']]['Ranges'][-1]['Subnets'][-1])

                        range_first_subnet = None
                        range_last_subnet = None
                        Logger.debug(f"{partition['name']} {i} excluded: {subnet}")
                        continue

                ### if start of a range, append a new Range
                if range_first_subnet is None:
                    range_first_subnet = subnet
                    result[partition['name']]['Ranges'].append({
                        'CIDR': range_first_subnet.with_prefixlen,
                        'Count': 0, ### to be incremented while adding subnets
                        'Subnets': []
                    })
                    if summary:
                        result[partition['name']]['Ranges'][-1]['Subnets'].append({
                            'CIDR': range_first_subnet.with_prefixlen,
                            'Name': partition['labels'].get(str(i), UNNAMED_SUBNET),
                        })
                        add_partition_subnets(partitions, partition_ix+1, range_first_subnet, result[partition['name']]['Ranges'][-1]['Subnets'][-1])
                else:
                    range_last_subnet = subnet
                    cidr = f"{range_first_subnet.with_prefixlen}-{range_last_subnet.with_prefixlen}"
                    ### updare Range CIDR
                    result[partition['name']]['Ranges'][-1]['CIDR'] = cidr
                    ### if in summary mode, always add the last subnet in the list
                    if summary and range_last_subnet == subnets[-1]:
                        result[partition['name']]['Ranges'][-1]['Subnets'].append({
                            'CIDR' : range_last_subnet.with_prefixlen,
                            'Name': partition['labels'].get(str(i), UNNAMED_SUBNET),
                        })
                        add_partition_subnets(partitions, partition_ix+1, range_last_subnet, result[partition['name']]['Ranges'][-1]['Subnets'][-1])

                ### always add the subnet if not in summary mode
                if not summary:
                    result[partition['name']]['Ranges'][-1]['Subnets'].append({
                        'CIDR' : subnet.with_prefixlen,
                        'Name': partition['labels'].get(str(i), UNNAMED_SUBNET),
                    })
                    add_partition_subnets(partitions, partition_ix+1, subnet, result[partition['name']]['Ranges'][-1]['Subnets'][-1])

                result[partition['name']]['Count'] = result[partition['name']]['Count'] + 1
                result[partition['name']]['Ranges'][-1]['Count'] = result[partition['name']]['Ranges'][-1]['Count'] + 1

            ### compact Range list if has only one element
            if len(result[partition['name']]['Ranges']) == 1:
                result[partition['name']].update(result[partition['name']]['Ranges'][0])
                result[partition['name']].pop('Ranges')

        ### skip partitions with 0 bits long, and move to the next partition
        else:
            add_partition_subnets(partitions, partition_ix+1, supernet, result)

    result = {
        'SubnetPlan': {
            'Name': subnet_plan['name'],
            'Description': subnet_plan['description'],
            'CIDR': subnet_plan['cidr']
        }
    }

    add_partition_subnets(subnet_plan['partitions'], 0, ipaddress.ip_network(subnet_plan['cidr']), result['SubnetPlan'])
    return result


subnet_plan_group = yaml.load(open('subnet-plan-group.yaml', 'r', encoding='utf-8'), yaml.SafeLoader)
for plan in subnet_plan_group['plans']:
    Logger.info(f"Executing subnet plan '{plan['name']}'...")
    planned = execute_subnet_plan(plan, summary=True)
    with open(f"{plan['name']}.yaml", 'w') as f:
        yaml.dump(planned, f, sort_keys=False, indent=2)



import boto3
from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers
from dsocli.configs import ConfigProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.dict_utils import set_dict_value
from dsocli.contexts import Contexts
from dsocli.aws_ssm_utils import *
from dsocli.appconfig import AppConfig


__default_spec = {
    'pathPrefix': '/dso/v1/config',
}

def get_default_spec():
    return __default_spec.copy()


class AwsSsmConfigProvider(ConfigProvider):

    def __init__(self):
        super().__init__('config/aws/ssm/v1')


    ### adds service name to the artifactStore prefix
    def get_path_prefix(self, service):
        configPathPrefix = AppConfig.config_spec('pathPrefix')
        if not configPathPrefix.endswith('/'): configPathPrefix += '/'
        return configPathPrefix + service


    def list(self, service, uninherited=False, filter=None):
        self.service = service
        Logger.debug(f"Listing configuration on SSM: service={service}")
        configuration = load_context_ssm_configuration(parameter_type='String', path_prefix=self.get_path_prefix(service), uninherited=uninherited, filter=filter)
        result = {'Parameters': []}
        for key, details in configuration.items():
            item = {
                'Key': key,
                'RevisionId': str(details['Version']),
            }
            item.update(details)
            result['Parameters'].append(item)

        return result


    def set(self, key, value, service):
        self.service = service
        Logger.debug(f"Checking configuration overwrites '{key}' on SSM: service={service}")
        assert_ssm_parameter_no_namespace_overwrites(key=key, path_prefix=self.get_path_prefix(service))
        Logger.debug(f"Locating configuration '{key}' on SSM: service={service}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(service), uninherited=True)
        if found and not found['Type'] == 'String':
            raise DSOException(f"Failed to add parameter '{key}' becasue becasue the key is not available in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        path = get_ssm_path(context=AppConfig.context, key=key, path_prefix=self.get_path_prefix(service))
        Logger.debug(f"Adding configuration: path={path}")
        response = add_ssm_paramater(path, value)
        result = {
                'RevisionId': str(response['Version']),
                'Key': key, 
                'Value': value,
                'Stage': AppConfig.short_stage,
                'Scope': AppConfig.context.scope_translation,
                'Origin': {
                    'Namespace': AppConfig.namespace,
                    'Project': AppConfig.project,
                    'Application': AppConfig.application,
                    'Stage': AppConfig.stage,
                },
                'Path': path,
            }
        result.update(response)
        return result



    def get(self, key, service, revision=None):
        self.service = service
        Logger.debug(f"Locating configuration '{key}' on SSM: service={service}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(service))
        if not found:
            raise DSOException(f"Config '{key}' not found nor inherited in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            if not found['Type'] == 'String':
                raise DSOException(f"Config '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Getting configuration from SSM: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        configuration = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if revision is None:
            ### get the latest revision
            result = {
                    'RevisionId': str(configuration[0]['Version']),
                    'Date': configuration[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': configuration[0]['Value'],
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'Path': found['Name'],
                    'User': configuration[0]['LastModifiedUser'],
                    }
                
        else:
            ### get specific revision
            configuration = [x for x in configuration if str(x['Version']) == revision]
            if not configuration:
                raise DSOException(f"Revision '{revision}' not found for parameter '{key}' in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
            result = {
                    'RevisionId':str(configuration[0]['Version']),
                    'Date': configuration[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': configuration[0]['Value'],
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'Path': found['Name'],
                    'User': configuration[0]['LastModifiedUser'],
                    }

        return result


    def history(self, key, service):
        self.service = service
        Logger.debug(f"Locating configuration '{key}' on SSM: service={service}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(service))
        if not found:
            raise DSOException(f"Config '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            if not found['Type'] == 'String':
                raise DSOException(f"Config '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Getting configuration from SSM: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        configuration = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        result = { "Revisions":
            [{
                'RevisionId': str(parameter['Version']),
                'Date': parameter['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                'Key': key,
                'Value': parameter['Value'],
                # 'Scope': found['Scope'],
                # 'Origin': found['Origin'],
                'User': parameter['LastModifiedUser'],
                # 'Path': found['Name'],
            } for parameter in configuration]
        }

        return result



    def unset(self, key, service):
        self.service = service
        Logger.debug(f"Locating configuration '{key}' on SSM: service={service}")
        ### only configuration owned by the context can be deleted, hence uninherited=True
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(service), uninherited=True)
        if not found:
            raise DSOException(f"Configuration '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            # if len(found) > 1:
            #     Logger.warn(f"More than one parameter found at '{found['Name']}'. The first one taken, and the rest were discarded.")
            if not found['Type'] == 'String':
                raise DSOException(f"Configuraton '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Deleting configuration from SSM: path={found['Name']}")
        delete_ssm_parameter(found['Name'])
        result = {
                'Key': key,
                'Stage': found['Stage'],
                'Scope': found['Scope'],
                'Origin': found['Origin'],
                'Path': found['Name'],
                }
        return result


def register():
    Providers.register(AwsSsmConfigProvider())




import os
import re
from .constants import *
from .providers import ProviderBase, Providers
from .artifacts import ArtifactStore
from .logger import Logger
from .exceptions import DSOException
from .stages import Stages
from .appconfig import AppConfig, ContextSource
from .configs import Configs


key_regex_pattern = r"^[a-zA-Z]([./a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class PackageProvider(ProviderBase):
    def list(self, filter=None):
        raise NotImplementedError()
    def build(self, key):
        raise NotImplementedError()
    def get(self, key, revision=None):
        raise NotImplementedError()
    # def history(self, key):
    #     raise NotImplementedError()
    def delete(self, key):
        raise NotImplementedError()


class PackageService():
    

    service_name = 'package'


    # def validate_key(self, key):
    #     Logger.info(f"Validating package key '{key}'...")
    #     if not key:
    #         raise DSOException(MESSAGES['KeyNull'])
    #     if key == 'dso' or key.startswith('dso.'):
    #         raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
    #     if not re.match(key_regex_pattern, key):
    #         raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, key_regex_pattern))
    #     ### the regex does not check adjacent special chars
    #     if '..' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))

    #     if '//' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '//'))


    @property
    def version_major(self):
        try:
            return int(Configs.get(key='version.major', service=self.service_name)['Value'])
        except DSOException:
            init_version_major = 0
            self.version_major = init_version_major
            return init_version_major

    @version_major.setter
    def version_major(self, value):
        Configs.set(key='version.major', value=str(value), service=self.service_name)


    @property
    def version_minor(self):
        try:
            return int(Configs.get(key='version.minor', service=self.service_name)['Value'])
        except DSOException:
            init_version_minor = 0
            self.version_minor = init_version_minor
            return init_version_minor

    @version_minor.setter
    def version_minor(self, value):
        Configs.set(key='version.minor', value=str(value), service=self.service_name)


    @property
    def version_patch(self):
        try:
            return int(Configs.get(key='version.patch', service=self.service_name)['Value'])
        except DSOException:
            init_version_patch = 0
            self.version_patch = init_version_patch
            return init_version_patch

    @version_patch.setter
    def version_patch(self, value):
        Configs.set(key='version.patch', value=str(value), service=self.service_name)


    @property
    def version_build(self):
        try:
            return int(Configs.get(key='version.build', service=self.service_name)['Value'])
        except DSOException:
            init_version_build = 0
            self.version_build = init_version_build
            return init_version_build


    @version_build.setter
    def version_build(self, value):
        Configs.set(key='version.build', value=str(value), service=self.service_name)


    def list(self, filter=None):
        provider = Providers.PackageProvider()
        Logger.info(f"Listing packages: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.list(service=self.service_name, filter=filter)
        result = {'Packages': response['Artifacts']}
        return result


    def build(self):
        provider = Providers.PackageProvider()
        Logger.info(f"Building package: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        artifact = provider.build()
        major = self.version_major
        minor = self.version_minor
        patch = self.version_patch
        build = self.version_build
        artifactKey = f"{major}.{minor}.{patch}.{build}"
        self.version_build = build + 1
        Logger.info(f"Adding package '{artifactKey}' to artifact store...")
        # changed = Logger.decrease_verbosity()
        # try:
        #     response = ArtifactStore.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        # finally:
        #     if changed: Logger.increase_verbosity()
        response = ArtifactStore.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        return response


    def get(self, key):
        Logger.info(f"Getting package '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.get(key=key, service=self.service_name)
        return response


    # def history(self, key):
    #     self.config = config
    #     provider = Providers.PackageProvider()
    #     Logger.info(f"Getting the history of package '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
    #     return provider.history(key)


    def delete(self, key):
        Logger.info(f"Deleting package '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.delete(key=key, service=self.service_name)
        return response


Packages = PackageService()



import os
import re
from .constants import *
from .providers import ProviderBase, Providers
from .artifacts import ArtifactStore
from .logger import Logger
from .exceptions import DSOException
from .stages import Stages
from .appconfig import AppConfig, ContextSource
from .configs import Configs

key_regex_pattern = r"^[a-zA-Z]([./a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class ReleaseProvider(ProviderBase):
    def list(self, filter=None):
        raise NotImplementedError()
    def build(self, key):
        raise NotImplementedError()
    def get(self, key, revision=None):
        raise NotImplementedError()
    # def history(self, key):
    #     raise NotImplementedError()
    def delete(self, key):
        raise NotImplementedError()


class ReleaseService():
    

    service_name = 'release'


    # def validate_key(self, key):
    #     Logger.info(f"Validating release key '{key}'...")
    #     if not key:
    #         raise DSOException(MESSAGES['KeyNull'])
    #     if key == 'dso' or key.startswith('dso.'):
    #         raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
    #     if not re.match(key_regex_pattern, key):
    #         raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, key_regex_pattern))
    #     ### the regex does not check adjacent special chars
    #     if '..' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))

    #     if '//' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '//'))


    @property
    def version_major(self):
        try:
            return int(Configs.get(key='version.major', service=self.service_name)['Value'])
        except DSOException:
            init_version_major = 0
            self.version_major = init_version_major
            return init_version_major

    @version_major.setter
    def version_major(self, value):
        Configs.set(key='version.major', value=str(value), service=self.service_name)


    @property
    def version_minor(self):
        try:
            return int(Configs.get(key='version.minor', service=self.service_name)['Value'])
        except DSOException:
            init_version_minor = 0
            self.version_minor = init_version_minor
            return init_version_minor

    @version_minor.setter
    def version_minor(self, value):
        Configs.set(key='version.minor', value=str(value), service=self.service_name)


    @property
    def version_patch(self):
        try:
            return int(Configs.get(key='version.patch', service=self.service_name)['Value'])
        except DSOException:
            init_version_patch = 0
            self.version_patch = init_version_patch
            return init_version_patch

    @version_patch.setter
    def version_patch(self, value):
        Configs.set(key='version.patch', value=str(value), service=self.service_name)


    @property
    def version_release(self):
        try:
            return int(Configs.get(key='version.release', service=self.service_name)['Value'])
        except DSOException:
            init_version_release = 0
            self.version_release = init_version_release
            return init_version_release


    @version_release.setter
    def version_release(self, value):
        Configs.set(key='version.release', value=str(value), service=self.service_name)


    def list(self, filter=None):
        provider = Providers.ReleaseProvider()
        Logger.info(f"Listing releases: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.list(service=self.service_name, filter=filter)
        result = {'Releases': response['Artifacts']}
        return result


    def create(self):
        provider = Providers.ReleaseProvider()
        Logger.info(f"Creating release: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        artifact = provider.create()
        major = self.version_major
        minor = self.version_minor
        patch = self.version_patch
        release = self.version_release
        artifactKey = f"{major}.{minor}.{patch}.{release}"
        self.version_release = release + 1
        Logger.info(f"Adding release '{artifactKey}' to artifact store...")
        # changed = Logger.decrease_verbosity()
        # try:
        #     response = ArtifactStore.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        # finally:
        #     if changed: Logger.increase_verbosity()
        response = ArtifactStore.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        return response


    def get(self, key):
        Logger.info(f"Getting release '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.get(key=key, service=self.service_name)
        return response


    # def history(self, key):
    #     self.config = config
    #     provider = Providers.ReleaseProvider()
    #     Logger.info(f"Getting the history of release '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
    #     return provider.history(key)


    def delete(self, key):
        Logger.info(f"Deleting release '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = ArtifactStore.delete(key=key, service=self.service_name)
        return response


Releases = ReleaseService()



import os
import re
from .constants import *
from .providers import Providers
from .logger import Logger
from .exceptions import DSOException
from .stages import Stages
from .appconfig import AppConfig, ContextSource


class ArtifactStoreService():

    def list(self, service, filter=None):
        provider = Providers.ArtifactStoreProvider()
        Logger.info(f"Listing artifacts: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        response = provider.list(filter=filter, service=service)
        result = {
            'Artifacts': response['Files']
        }
        return result

    def add(self, filepath, service, key=None):
        if not key: key = os.path.basename(filepath)
        # self.validate_key(key)
        provider = Providers.ArtifactStoreProvider()
        Logger.info(f"Adding artifact '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.add(filepath=filepath, key=key, service=service)


    def get(self, key, service):
        provider = Providers.ArtifactStoreProvider()
        Logger.info(f"Getting artifact '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.get(key=key, service=service)


    # def history(self, key):
    #     self.AppConfig = AppConfig
    #     provider = Providers.ArtifactStoreProvider()
    #     Logger.info(f"Getting the history of artifact '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
    #     return provider.history(key)


    def delete(self, key, service):
        provider = Providers.ArtifactStoreProvider()
        Logger.info(f"Deleting artifact '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        return provider.delete(key=key, service=service)



ArtifactStore = ArtifactStoreService()



import re
import boto3
from botocore.exceptions import ClientError
from .contexts import Contexts, Context
from .logger import Logger
from .stages import Stages
import logging
from .file_utils import *
from dsocli.appconfig import AppConfig

logging.getLogger('botocore').setLevel(Logger.mapped_level)
logging.getLogger('boto').setLevel(Logger.mapped_level)
logging.getLogger('boto3').setLevel(Logger.mapped_level)

UPLOAD_MULTIPART_THRESHOLD = 1024*25
UPLOAD_MULTIPART_CHUNK_SIZE = 1024*25
UPLOAD_MULTIPART_CONCURRENCY = 10

def get_s3_path(context, path_prefix='', key=None):
    if path_prefix.endswith('/'): path_prefix = path_prefix[:-1]
    return path_prefix + context.get_path(key)



# def list_s3_objects(bucket, path=''):
#     s3 = boto3.session.Session().client(service_name='s3')
#     paginator = s3.get_paginator("list_objects_v2")
#     response = paginator.paginate(Bucket=bucket, Prefix=path).build_full_result()
#     return response.get('Contents', [])


# def load_s3_path(result, bucket, path, filter=None, used_path_prefix='', object_kind='object'):
#     archives = list_s3_objects(bucket, path=path)
#     for archive in archives:
#         key = archive['Key'][len(path)+1:]
#         if filter and not re.match(filter, key): continue
#         if key in result:
#             Logger.warn(f"Inherited {object_kind} '{key}' was overridden.")
#         ctx_path = path[len(used_path_prefix):]
#         ctx = Context(*Contexts.parse_path(ctx_path)[0:4])
#         details = {
#                     'Bucket': bucket,
#                     'Path': archive['Key'],
#                     'ETag': archive['ETag'],
#                     'Stage': ctx.short_stage,
#                     'Scope': ctx.scope_translation,
#                     'Origin': {
#                         'Namespace': ctx.namespace,
#                         'Project': ctx.project,
#                         'Application': ctx.application,
#                         'Stage': ctx.stage,
#                     },
#                     'Size': convert_file_size_unit(archive['Size']),
#                     # 'Version': archive['Version'],
#                     'Date': archive['LastModified'].strftime('%Y/%m/%d-%H:%M:%S'),
#                 }
#         result[key] = details
#     return result



# def s3_context_list_files(bucket, path_prefix='', uninherited=False, filter=None, object_kind='object'):
#     ### construct search path in hierachy with no key specified
#     paths = Contexts.get_hierachy_paths(context=AppConfig.context, key=None, path_prefix=path_prefix, ignore_stage=AppConfig.stage is None, uninherited=uninherited)
#     archives = {}
#     for path in paths:
#         Logger.debug(f"Loading S3 path: bucket={bucket}, path={path}")
#         load_s3_path(config=bucket=bucket, result=archives, path=path, filter=filter, used_path_prefix=path_prefix, object_kind=object_kind)
#     return archives



def list_s3_objects(bucket, path, filter=None):
    s3 = boto3.session.Session().client(service_name='s3')
    paginator = s3.get_paginator("list_objects_v2")
    response = paginator.paginate(Bucket=bucket, Prefix=path).build_full_result()
    items = response.get('Contents', [])
    result = []
    for item in items:
        key = item['Key'][len(path)+1:]
        if filter and not re.match(filter, key): continue
        details = {
                    'Key': key,
                    'Path': item['Key'],
                    'ETag': item['ETag'],
                    'Size': convert_file_size_unit(item['Size']),
                    # 'Version': item['Version'],
                    'Date': item['LastModified'].strftime('%Y/%m/%d-%H:%M:%S'),
                }
        result.append(details)
    return result



def s3_context_list_files(bucket, path_prefix='', filter=None):
    path = get_s3_path(AppConfig.context, path_prefix, key=None)
    Logger.debug(f"Loading S3 path: bucket={bucket}, path={path}")
    response = list_s3_objects(bucket, path, filter)
    return response



def s3_context_add_file(filepath, bucket, key, path_prefix=''):
    s3 = boto3.session.Session().client(service_name='s3')
    transferConfig = boto3.s3.transfer.TransferConfig(multipart_threshold=UPLOAD_MULTIPART_THRESHOLD, max_concurrency=UPLOAD_MULTIPART_CONCURRENCY, multipart_chunksize=UPLOAD_MULTIPART_CHUNK_SIZE, use_threads=True)
    path = get_s3_path(AppConfig.context, path_prefix, key)
    Logger.debug(f"Uploading S3 object: bucket={bucket}, path={path}")
    s3.upload_file(filepath, bucket, path, Config=transferConfig)
    result = {
        'Bucket': bucket,
        'Path': path,
    }
    return result



def s3_context_get_file(bucket, key, path_prefix=''):
    s3 = boto3.session.Session().client(service_name='s3')
    path = get_s3_path(AppConfig.context, path_prefix, key)
    Logger.debug(f"Downloading S3 object: bucket={bucket}, path={path}")
    try:
        s3.download_file(Bucket=bucket, Key=path, Filename=key)
    except ClientError as e:
        if 'Not Found' in repr(e):
            return {}
        else:
            raise

    result = {
        'Key': key,
        'Bucket': bucket,
        'Path': path,
        'FilePath': '.' + os.path.join(os.sep, key)
    }
    return result


def s3_context_delete_file(bucket, key, path_prefix=''):
    s3 = boto3.session.Session().client(service_name='s3')
    path = get_s3_path(AppConfig.context, path_prefix, key)
    Logger.debug(f"Deleting S3 object: bucket={bucket}, path={path}")
    try:
        s3.head_object(Bucket=bucket, Key=path)
    except ClientError as e:
        if 'Not Found' in repr(e):
            return {}
        else:
            raise

    s3.delete_object(Bucket=bucket, Key=path)
    result = {
        'Key': key,
        'Bucket': bucket,
        'Path': path,
    }
    return result




import boto3
from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers
from dsocli.secrets import SecretProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.dict_utils import set_dict_value
from dsocli.contexts import Contexts
from dsocli.aws_ssm_utils import *
from dsocli.appconfig import AppConfig


__default_spec = {
    'pathPrefix': '/dso/v1/secret/',
    # 'namnespace': 'default',
}

def get_default_spec():
    return __default_spec.copy()


class AwsSsmSecretProvider(SecretProvider):


    def __init__(self):
        super().__init__('secret/aws/ssm/v1')


    def get_path_prefix(self):
        return AppConfig.secret_spec('pathPrefix')


    def list(self, uninherited=False, decrypt=False, filter=None):
        Logger.debug(f"Listing SSM secrets: namesape={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        secrets = load_context_ssm_parameters(parameter_type='SecureString', path_prefix=self.get_path_prefix(), uninherited=uninherited, decrypt=decrypt, filter=filter)
        result = {'Secrets': []}
        for key, details in secrets.items():
            item = {
                'Key': key,
                'RevisionId': str(details['Version']),
            }
            item.update(details)
            result['Secrets'].append(item)

        return result



    def add(self, key, value):
        Logger.debug(f"Checking SSM secret overwrites '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        assert_ssm_parameter_no_namespace_overwrites(key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Locating SSM secret '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if found and not found['Type'] == 'SecureString':
            raise DSOException(f"Failed to add secret '{key}' becasue becasue the key is not available in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        path = get_ssm_path(context=AppConfig.context, key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Adding SSM secret: path={path}")
        response = add_ssm_secret(path, value)
        result = {
                'RevisionId': str(response['Version']),
                'Key': key, 
                'Value': value,
                'Stage': AppConfig.short_stage,
                'Scope': AppConfig.context.scope_translation,
                'Origin': {
                    'Namespace': AppConfig.namespace,
                    'Project': AppConfig.project,
                    'Application': AppConfig.application,
                    'Stage': AppConfig.stage,
                },
                'Path': path,
            }
        result.update(response)
        return result


    def get(self, key, decrypt=False, revision=None, uninherited=False, editing=False):
        Logger.debug(f"Locating SSM secret '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=uninherited)
        if not found:
            if uninherited:
                raise DSOException(f"Secret '{key}' not found in the given context: namesape={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
            else:
                raise DSOException(f"Secret '{key}' not found nor inherited in the given context: namesape={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        else:
            if not found['Type'] == 'SecureString':
                raise DSOException(f"Secret '{key}' not found in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Getting SSM secret: path={found['Name']}")
        response = get_ssm_secret_history(found['Name'], decrypt)
        secrets = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if revision is None:
            ### get the latest revision
            result = {
                    'RevisionId': str(secrets[0]['Version']),
                    'Date': secrets[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': secrets[0]['Value'],
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'Path': found['Name'],
                    'User': secrets[0]['LastModifiedUser'],
                    }

        else:
            ### get specific revision
            secrets = [x for x in secrets if str(x['Version']) == revision]
            if not secrets:
                raise DSOException(f"Revision '{revision}' not found for secret '{key}' in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
            result = {
                    'RevisionId': str(secrets[0]['Version']),
                    'Date': secrets[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': secrets[0]['Value'],
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'Path': found['Name'],
                    'User': secrets[0]['LastModifiedUser'],
                    }

        return result



    def history(self, key, decrypt=False):
        Logger.debug(f"Locating SSM secret '{key}': project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix())
        if not found:
            raise DSOException(f"Secret '{key}' not found in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            if not found['Type'] == 'SecureString':
                raise DSOException(f"Parameter '{key}' not found in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Getting SSM secret: path={found['Name']}")
        response = get_ssm_secret_history(found['Name'], decrypt)
        secrets = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        result = { "Revisions":
            [{
                'RevisionId': str(secret['Version']),
                'Date': secret['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                'Key': key,
                'Value': secret['Value'],
                # 'Scope': found['Scope'],
                # 'Origin': found['Origin'],
                'User': secret['LastModifiedUser'],
                # 'Path': found['Name'],
            } for secret in secrets]
        }

        return result



    def delete(self, key):
        Logger.debug(f"Locating SSM secret '{key}': namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        ### only secrets owned by the context can be deleted, hence uninherited=True
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
                raise DSOException(f"Secret '{key}' not found in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            # if len(found) > 1:
            #     Logger.warn(f"More than one secret found at '{found['Name']}'. The first one taken, and the rest were discarded.")
            if not found['Type'] == 'SecureString':
                raise DSOException(f"Secret '{key}' not found in the given context: namespace={AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Deleting SSM secret: path={found['Name']}")
        delete_ssm_parameter(found['Name'])
        return {
                'Key': key,
                'Stage': found['Stage'],
                'Scope': found['Scope'],
                'Origin': found['Origin'],
                'Path': found['Name'],
                }



def register():
    Providers.register(AwsSsmSecretProvider())



import os
import re
from .contexts import Context, Contexts
from .logger import Logger
from .stages import Stages
from dsocli.file_utils import *
from dsocli.exceptions import DSOException
from dsocli.dict_utils import *
from dsocli.local_utils import *
from dsocli.appconfig import AppConfig

def expand_shell_parameters(parameters):
    for param in parameters:
        parameters[param]['Value'] = os.getenv(parameters[param]['Value'], '')
    return parameters


def locate_shell_parameter_in_context_hierachy(key, store_name, path_prefix='', uninherited=False, raw=False):
    parameters = locate_parameter_in_context_hierachy(key=key, store_name=store_name, path_prefix=path_prefix, uninherited=uninherited)
    if raw:
        return parameters
    else:
        return expand_shell_parameters(parameters)

def add_shell_parameter(key, value, store_name, path_prefix=''):
    return add_local_parameter(key=key, value=value, store_name=store_name, path_prefix=path_prefix)

def delete_shell_parameter(path, key):
    return delete_local_parameter(path=path, key=key)

def load_context_shell_parameters(store_name, path_prefix='', uninherited=False, filter=None, raw=False):
    parameters = load_context_local_parameters(store_name=store_name, path_prefix=path_prefix, uninherited=uninherited, filter=filter)
    if raw:
        return parameters
    else:
        return expand_shell_parameters(parameters)





import re

from jinja2.runtime import unicode_join
from .logger import Logger
from .providers import KeyValueStoreProvider, Providers
from .stages import Stages
from .constants import *
from .exceptions import *
from .appconfig import AppConfig, ContextSource


key_regex_pattern = r"^[a-zA-Z]([.a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class ParameterProvider(KeyValueStoreProvider):
    def list(self, uninherited=False, filter=None):
        raise NotImplementedError()
    def add(self, key, value):
        raise NotImplementedError()
    def get(self, key, revision=None):
        raise NotImplementedError()
    def history(self, key):
        raise NotImplementedError()
    def delete(self, key):
        raise NotImplementedError()


class ParameterService():

    def validate_key(self, key):
        Logger.info(f"Validating parameter key '{key}'...")
        if not key:
            raise DSOException(MESSAGES['KeyNull'])
        if key == 'dso' or key.startswith('dso.'):
            raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
        if not re.match(key_regex_pattern, key):
            raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, key_regex_pattern))
        if '..' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))
            
    def list(self, uninherited=False, filter=None):
        provider = Providers.ParameterProvider()
        Logger.info(f"Listing parameters: namespace={AppConfig.get_namespace(ContextSource.Target)}, project={AppConfig.get_project(ContextSource.Target)}, application={AppConfig.get_application(ContextSource.Target)}, stage={AppConfig.get_stage(ContextSource.Target, short=True)}, scope={AppConfig.scope}")
        return provider.list(uninherited, filter)

    def add(self, key, value):
        self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Adding parameter '{key}': namespace={AppConfig.get_namespace(ContextSource.Target)}, project={AppConfig.get_project(ContextSource.Target)}, application={AppConfig.get_application(ContextSource.Target)}, stage={AppConfig.get_stage(ContextSource.Target, short=True)}, scope={AppConfig.scope}")
        return provider.add(key, value)

    def get(self, key, revision=None, uninherited=False, editable=False):
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Getting parameter '{key}': namespace={AppConfig.get_namespace(ContextSource.Target)}, project={AppConfig.get_project(ContextSource.Target)}, application={AppConfig.get_application(ContextSource.Target)}, stage={AppConfig.get_stage(ContextSource.Target, short=True)}, scope={AppConfig.scope}")
        return provider.get(key, revision, uninherited, editable)

    def history(self, key):
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Getting the history of parameter '{key}': namespace={AppConfig.get_namespace(ContextSource.Target)}, project={AppConfig.get_project(ContextSource.Target)}, application={AppConfig.get_application(ContextSource.Target)}, stage={AppConfig.get_stage(ContextSource.Target, short=True)}, scope={AppConfig.scope}")
        return provider.history(key)

    def delete(self, key):
        # self.validate_key(key)
        provider = Providers.ParameterProvider()
        Logger.info(f"Deleting parameter '{key}': namespace={AppConfig.get_namespace(ContextSource.Target)}, project={AppConfig.get_project(ContextSource.Target)}, application={AppConfig.get_application(ContextSource.Target)}, stage={AppConfig.get_stage(ContextSource.Target, short=True)}, scope={AppConfig.scope}")
        return provider.delete(key)

Parameters = ParameterService()


from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers
from dsocli.artifacts import ArtifactStore
from dsocli.packages import PackageProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.contexts import Contexts
from dsocli.aws_s3_utils import *
from dsocli.settings import *
from dsocli.appconfig import AppConfig

__default_spec = {
    'foo': 'bar',
}


def get_default_spec():
    return __default_spec.copy()


class AwsCfnPackageProvider(PackageProvider):

    def __init__(self):
        super().__init__('package/aws/cfn/v1')

    def build(self):
        Logger.debug(f"Building CFN package: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.stage}")
        return 'package1.zip'


def register():
    Providers.register(AwsCfnPackageProvider())



from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers, ArtifactStoreProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.contexts import Contexts
from dsocli.aws_s3_utils import *
from dsocli.settings import *
from dsocli.appconfig import AppConfig


__default_spec = {
    'bucket': 'mybucket',
    'pathPrefix': 'dso/v1/',
}


def get_default_spec():
    return __default_spec.copy()


class AwsS3ArtifactStoreProvider(ArtifactStoreProvider):

    def __init__(self):
        super().__init__('artifactStore/aws/s3/v1')


    def get_bucket_name(self):
        return AppConfig.artifactStore_spec('bucket')

    ### adds service name to the artifactStore prefix
    def get_path_prefix(self, service):
        storePathPrefix = AppConfig.artifactStore_spec('pathPrefix')
        if not storePathPrefix.endswith('/'): storePathPrefix += '/'
        return storePathPrefix + service

    def list(self, service, filter=None):
        Logger.debug(f"Listing artifacts from S3: bucket={self.get_bucket_name()}")
        items = s3_context_list_files(bucket=self.get_bucket_name(), path_prefix=self.get_path_prefix(service), filter=filter)
        result = {'Files': items}
        return result


    def add(self, filepath, key, service):
        Logger.debug(f"Adding artifact '{key}' to S3: bucket={self.get_bucket_name()}")
        response = s3_context_add_file(filepath=filepath, bucket=self.get_bucket_name(), key=key, path_prefix=self.get_path_prefix(service))
        result = {
                # 'RevisionId': str(response['Version']),
                'Key': key,
                'Context': {
                    'Namespace': AppConfig.namespace,
                    'Project': AppConfig.project,
                    'Application': AppConfig.application,
                    'Stage': AppConfig.stage,
                },
            }
        result.update(response)
        return result


    def get(self, key, service, revision=None):
        Logger.debug(f"Getting artifact '{key}' from S3: bucket={self.get_bucket_name()}")
        response = s3_context_get_file(bucket=self.get_bucket_name(), key=key, path_prefix=self.get_path_prefix(service))
        if not response:
            raise DSOException(f"Artifact '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        result = {
                # 'RevisionId': str(response['Version']),
                'Key': key,
                'Context': {
                    'Namespace': AppConfig.namespace,
                    'Project': AppConfig.project,
                    'Application': AppConfig.application,
                    'Stage': AppConfig.stage,
                },
            }
        result.update(response)
        return result

        # response = get_ssm_artifact_history(found['Name'])
        # artifact = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        # if revision is None:
        #     ### get the latest revision
        #     result = {
        #             'RevisionId': str(artifact[0]['Version']),
        #             'Date': artifact[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
        #             'Key': key, 
        #             'Scope': found['Scope'],
        #             'Origin': found['Origin'],
        #             'User': artifact[0]['LastModifiedUser'],
        #             'Path': found['Name'],
        #             'Contents': artifact[0]['Value'],
        #             }
        # else:
        #     ### get specific revision
        #     artifact = [x for x in artifact if str(x['Version']) == revision]
        #     if not artifact:
        #         raise DSOException(f"Revision '{revision}' not found for artifact '{key}' in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        #     result = {
        #             'RevisionId': str(artifact[0]['Version']),
        #             'Date': artifact[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
        #             'Key': key, 
        #             'Scope': found['Scope'],
        #             'Origin': found['Origin'],
        #             'Path': found['Name'],
        #             'User': artifact[0]['LastModifiedUser'],
        #             'Contents': artifact[0]['Value'],
        #             }

        return result



    def history(self, key, include_contents=False):
        Logger.debug(f"Locating artifact '{key}' from S3: bucket={self.get_bucket_name()}")
        found = locate_ssm_parameter_in_context_hierachy(key=key, service=self.get_path_prefix())
        if not found:
            raise DSOException(f"Artifact '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        else:
            if not found['Type'] == 'StringList':
                raise DSOException(f"Storage '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        Logger.debug(f"Getting S3 artifact: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        artifact = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if include_contents:
            result = { "Revisions":
                [{
                    'RevisionId': str(artifact['Version']),
                    'Date': artifact['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': artifact['LastModifiedUser'],
                    # 'Path': found['Name'],
                    'Contents': artifact[0]['Value'],

                } for artifact in artifact]
            }
        else:
            result = { "Revisions":
                [{
                    'RevisionId': str(artifact['Version']),
                    'Date': artifact['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': artifact['LastModifiedUser'],
                    # 'Path': found['Name'],
                } for artifact in artifact]
            }

        return result



    def delete(self, key, service):
        Logger.debug(f"Deleting artifact '{key}' from S3: bucket={self.get_bucket_name()}")
        response = s3_context_delete_file(bucket=self.get_bucket_name(), key=key, path_prefix=self.get_path_prefix(service))
        if not response:
            raise DSOException(f"Artifact '{key}' not found in the given context: namespace:{AppConfig.namespace}, project={AppConfig.project}, application={AppConfig.application}, stage={AppConfig.short_stage}")
        result = {
                # 'RevisionId': str(response['Version']),
                'Key': key,
                'Context': {
                    'Namespace': AppConfig.namespace,
                    'Project': AppConfig.project,
                    'Application': AppConfig.application,
                    'Stage': AppConfig.stage,
                },
            }
        result.update(response)
        return result



def register():
    Providers.register(AwsS3ArtifactStoreProvider())
